==> symbol-processing-aa-embeddable-2.1.0-1.0.29-sources.jar: org/jetbrains/kotlin/fir/analysis/checkers/declaration/FirPropertyInitializationChecker.kt <==
/*
 * Copyright 2010-2021 JetBrains s.r.o. and Kotlin Programming Language contributors.
 * Use of this source code is governed by the Apache 2.0 license that can be found in the license/LICENSE.txt file.
 */

package ksp.org.jetbrains.kotlin.fir.analysis.checkers.declaration

import ksp.org.jetbrains.kotlin.diagnostics.DiagnosticReporter
import ksp.org.jetbrains.kotlin.diagnostics.reportOn
import ksp.org.jetbrains.kotlin.fir.FirElement
import ksp.org.jetbrains.kotlin.fir.analysis.checkers.MppCheckerKind
import ksp.org.jetbrains.kotlin.fir.analysis.checkers.context.CheckerContext
import ksp.org.jetbrains.kotlin.fir.analysis.diagnostics.FirErrors
import ksp.org.jetbrains.kotlin.fir.declarations.*
import ksp.org.jetbrains.kotlin.fir.expressions.FirVariableAssignment
import ksp.org.jetbrains.kotlin.fir.expressions.calleeReference
import ksp.org.jetbrains.kotlin.fir.references.toResolvedCallableSymbol
import ksp.org.jetbrains.kotlin.fir.symbols.impl.FirPropertySymbol
import ksp.org.jetbrains.kotlin.fir.visitors.FirVisitorVoid

object FirPropertyInitializationChecker : FirRegularClassChecker(MppCheckerKind.Common) {
    override fun check(declaration: FirRegularClass, context: CheckerContext, reporter: DiagnosticReporter) {
        val declaredLater = mutableSetOf<FirPropertySymbol>()
        val visitor = object : FirVisitorVoid() {
            override fun visitElement(element: FirElement) = element.acceptChildren(this)

            override fun visitConstructor(constructor: FirConstructor) {}

            override fun visitSimpleFunction(simpleFunction: FirSimpleFunction) {}

            override fun visitPropertyAccessor(propertyAccessor: FirPropertyAccessor) {}

            override fun visitAnonymousFunction(anonymousFunction: FirAnonymousFunction) {
                // TODO? if (anonymousFunction.invocationKind.isInPlace) visitElement(anonymousFunction)
            }

            override fun visitRegularClass(regularClass: FirRegularClass) {}

            override fun visitEnumEntry(enumEntry: FirEnumEntry) {}

            override fun visitVariableAssignment(variableAssignment: FirVariableAssignment) {
                variableAssignment.acceptChildren(this)
                val propertySymbol = variableAssignment.calleeReference?.toResolvedCallableSymbol() as? FirPropertySymbol ?: return
                if (propertySymbol !in declaredLater) return
                reporter.reportOn(variableAssignment.lValue.source, FirErrors.INITIALIZATION_BEFORE_DECLARATION, propertySymbol, context)
            }
        }

        for (member in declaration.declarations.asReversed()) {
            if (declaredLater.isNotEmpty()) {
                member.accept(visitor)
            }
            if (member is FirProperty) {
                declaredLater.add(member.symbol)
            }
        }
    }
}

==> symbol-processing-aa-embeddable-2.2.0-2.0.2-sources.jar: org/jetbrains/kotlin/fir/analysis/checkers/declaration/FirPropertyInitializationChecker.kt <==
/*
 * Copyright 2010-2021 JetBrains s.r.o. and Kotlin Programming Language contributors.
 * Use of this source code is governed by the Apache 2.0 license that can be found in the license/LICENSE.txt file.
 */

package ksp.org.jetbrains.kotlin.fir.analysis.checkers.declaration

import ksp.org.jetbrains.kotlin.diagnostics.DiagnosticReporter
import ksp.org.jetbrains.kotlin.diagnostics.reportOn
import ksp.org.jetbrains.kotlin.fir.FirElement
import ksp.org.jetbrains.kotlin.fir.analysis.checkers.MppCheckerKind
import ksp.org.jetbrains.kotlin.fir.analysis.checkers.context.CheckerContext
import ksp.org.jetbrains.kotlin.fir.analysis.diagnostics.FirErrors
import ksp.org.jetbrains.kotlin.fir.declarations.*
import ksp.org.jetbrains.kotlin.fir.expressions.FirVariableAssignment
import ksp.org.jetbrains.kotlin.fir.expressions.calleeReference
import ksp.org.jetbrains.kotlin.fir.references.toResolvedCallableSymbol
import ksp.org.jetbrains.kotlin.fir.symbols.impl.FirPropertySymbol
import ksp.org.jetbrains.kotlin.fir.visitors.FirVisitorVoid

object FirPropertyInitializationChecker : FirRegularClassChecker(MppCheckerKind.Common) {
    context(context: CheckerContext, reporter: DiagnosticReporter)
    override fun check(declaration: FirRegularClass) {
        val declaredLater = mutableSetOf<FirPropertySymbol>()
        val visitor = object : FirVisitorVoid() {
            override fun visitElement(element: FirElement) = element.acceptChildren(this)

            override fun visitConstructor(constructor: FirConstructor) {}

            override fun visitSimpleFunction(simpleFunction: FirSimpleFunction) {}

            override fun visitPropertyAccessor(propertyAccessor: FirPropertyAccessor) {}

            override fun visitAnonymousFunction(anonymousFunction: FirAnonymousFunction) {
                // TODO? if (anonymousFunction.invocationKind.isInPlace) visitElement(anonymousFunction)
            }

            override fun visitRegularClass(regularClass: FirRegularClass) {}

            override fun visitEnumEntry(enumEntry: FirEnumEntry) {}

            override fun visitVariableAssignment(variableAssignment: FirVariableAssignment) {
                variableAssignment.acceptChildren(this)
                val propertySymbol = variableAssignment.calleeReference?.toResolvedCallableSymbol() as? FirPropertySymbol ?: return
                if (propertySymbol !in declaredLater) return
                reporter.reportOn(variableAssignment.lValue.source, FirErrors.INITIALIZATION_BEFORE_DECLARATION, propertySymbol)
            }
        }

        @OptIn(DirectDeclarationsAccess::class)
        for (member in declaration.declarations.asReversed()) {
            if (declaredLater.isNotEmpty()) {
                member.accept(visitor)
            }
            if (member is FirProperty) {
                declaredLater.add(member.symbol)
            }
        }
    }
}
